@@ -4,10 +4,8 @@
 ;; A common input format for LaTeX and Html documents
 ;; This file realizes the translation to Html format.
 ;;
-;;     $Modified: Mon Sep 15 15:54:02 1997 by otfried $
-;;
 ;; This file is part of Hyperlatex
-;; Copyright (C) 1994, 1995, 1996, 1997 Otfried Schwarzkopf	
+;; Copyright (C) 1994-1998 Otfried Cheong	
 ;;  
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -47,7 +45,7 @@
 ;; set to true to see all macro expansions
 (setq hyperlatex-show-expansions nil)
 
-(defvar hyperlatex-version "2.2.6")
+(defvar hyperlatex-version "2.3.1")
 
 ;;------------------------------------------------------------------------
 
@@ -61,12 +59,28 @@
 (if (fboundp 'parse-colon-path)
     ()
   (defun parse-colon-path (arg)
-    arg))
+    (list (concat arg "/"))))
      
 (if (fboundp 'string-to-number)
     ()
   (fset 'string-to-number 'string-to-int)
   (fset 'number-to-string 'int-to-string))
+
+(if (fboundp 'find-file-literally)
+    ()
+  (fset 'find-file-literally 'find-file)) 
+
+(if (fboundp 'insert-file-contents-literally)
+    ()
+  (fset 'insert-file-contents-literally 'insert-file-contents))
+
+;; for Emacs 18 and 19
+(defvar enable-multibyte-characters)
+
+(if (fboundp 'set-buffer-multibyte)
+    ()
+  (defun set-buffer-multibyte (arg)
+    (setq enable-multibyte-characters arg)))
      
 (defvar	hyperlatex-produced-from)
 (defvar hyperlatex-active-space)
@@ -108,6 +122,7 @@
 (defvar hyperlatex-node-names)
 (defvar hyperlatex-node-number)
 (defvar hyperlatex-node-section)
+(defvar hyperlatex-document-options)
 (defvar hyperlatex-options)
 (defvar hyperlatex-packages)
 (defvar hyperlatex-produced-from-file)
@@ -119,23 +134,23 @@
 (defvar hyperlatex-special-chars-regexp)
 (defvar hyperlatex-stack)
 (defvar hyperlatex-tabular-column-descr)
+(defvar hyperlatex-tabular-column-types)
 (defvar hyperlatex-title)
+(defvar hyperlatex-topnode-name)
 
 (defvar hyperlatex-html-levels
   '(("html2" 20) ("html3.2" 32)))
 
 (defvar hyperlatex-format-syntax-table nil)
 
+;; File extension for HTML files.
+;; Change this to .htm for braindamaged systems
+(defvar hyperlatex-html-ext ".html")
+
 (defvar hyperlatex-extension-dirs
   (cons "~/.hyperlatex/" (parse-colon-path (getenv "HYPERLATEX_DIR"))))
 
-(defvar hyperlatex-is-mule (boundp 'mule-version)
-  "True if running under mule." )
-
-;; definition needed if not byte-compiled under Mule
-(defvar file-coding-system)
-
-(setq hyperlatex-meta-offset (if hyperlatex-is-mule 2176 128))
+(setq hyperlatex-meta-offset 128)
 
 (defun hyperlatex-meta (ch)
   (char-to-string (+ ch hyperlatex-meta-offset)))
@@ -158,7 +173,7 @@
 (setq hyperlatex-a-name-format
       (concat "A NAME=" hyperlatex-meta-dq "%s" hyperlatex-meta-dq))
 (setq hyperlatex-a-href-format
-      (concat "A HREF=" hyperlatex-meta-dq "%s" hyperlatex-meta-dq))
+      (concat "%s HREF=" hyperlatex-meta-dq "%s" hyperlatex-meta-dq))
 
 ;; all characters that need to be protected before working on the file
 (setq hyperlatex-meta-protect
@@ -249,13 +264,13 @@ Each file will be processed even if an error occurred previously."
   (let ((auto-save-default nil)
 	(find-file-run-dired nil)
 	(error 0)
-	(file-coding-system-for-read '*iso-8859-1*)
 	(file (expand-file-name (car command-line-args-left))))
     (if (not (file-exists-p file))
 	(error ">> %s does not exist!" file))
-    (find-file file)
+    (find-file-literally file)
     (buffer-disable-undo (current-buffer))
     (message "Hyperlatex formatting %s..." file)
+    (message "  (on Emacs %s)" emacs-version)
     (kill-emacs (hyperlatex-format-buffer-0))))
       
 (defun hyperlatex-format-buffer ()
@@ -263,16 +278,16 @@ Each file will be processed even if an error occurred previously."
 The Html file output is generated in a directory specified in the
  \\htmldirectory command, or in the current directory."
   (interactive)
-  (setq hyperlatex-message-buffer (get-buffer-create "*Hyperlatex messages*"))
+  (setq hyperlatex-message-buffer
+	(get-buffer-create "*Hyperlatex messages*"))
   (save-excursion
     (set-buffer hyperlatex-message-buffer)
     (buffer-disable-undo hyperlatex-message-buffer)
     (delete-region (point-min) (point-max)))
   (setq hyperlatex-message-window (display-buffer hyperlatex-message-buffer))
-  (let ((standard-output hyperlatex-message-buffer))
-    (if (zerop (hyperlatex-format-buffer-0))
-	(message "Hyperlatex formatting done.")
-      (error "Hyperlatex formatting error."))))
+  (if (zerop (hyperlatex-format-buffer-0))
+      (message "Hyperlatex formatting done.")
+    (error "Hyperlatex formatting error.")))
 
 (defun hyperlatex-format-buffer-0 ()
   "Process the current buffer as hyperlatex code.
@@ -284,6 +299,10 @@ Returns 0 if no error, 1 otherwise."
 	(hyperlatex-message "Hyperlatex formatting done.")
 	0)
     (error
+     (setq standard-output
+	   (if noninteractive
+	       t
+	     hyperlatex-message-buffer))
      (princ "\nHyperlatex ERROR: ")
      (if (eq (car err) 'error)
 	 (princ (eval (cons 'format (cdr err))))
@@ -300,6 +319,7 @@ Returns 0 if no error, 1 otherwise."
 (defun hyperlatex-message (form &optional arg1 arg2 arg3)
   (if noninteractive
       (message form arg1 arg2 arg3)
+    (setq standard-output hyperlatex-message-buffer)
     (princ (format form arg1 arg2 arg3))
     (princ "\n")
     (set-window-point hyperlatex-message-window
@@ -344,8 +364,7 @@ Returns 0 if no error, 1 otherwise."
 	 (hyperlatex-input-directory default-directory))
     ;;----------------------------------------------------------
     (set-buffer (get-buffer-create " *Hyperlatex Html output*"))
-    (if hyperlatex-is-mule
-	(setq file-coding-system '*iso-8859-1*))
+    (set-buffer-multibyte nil)
     (fundamental-mode)
     (set-syntax-table hyperlatex-format-syntax-table)
     (setq case-fold-search nil)
@@ -370,13 +389,15 @@ second pass."
 	(hyperlatex-active-space nil)
 	(hyperlatex-current-ref "")
 	(hyperlatex-tabular-column-descr nil)
+	(hyperlatex-tabular-column-types nil)
 	(hyperlatex-html-accents nil)
 	(hyperlatex-attributes nil)
 	(hyperlatex-new-commands nil)
 	(hyperlatex-make-panel t)
 	(hyperlatex-made-panel nil)
 	(hyperlatex-math-mode nil)
-	(hyperlatex-options "")
+	(hyperlatex-document-options nil)
+	(hyperlatex-options nil)
 	(hyperlatex-footnotes nil)
 	(hyperlatex-counters nil)
 	(hyperlatex-footnote-number 0)
@@ -418,7 +439,7 @@ second pass."
 			  hyperlatex-meta-})))
     (goto-char (point-min))
     (while (re-search-forward hyperlatex-meta-protect nil t)
-      (replace-match (format meta-iso-format (preceding-char))))))
+      (replace-match (format meta-iso-format (preceding-char)) t))))
 
 (defun hyperlatex-final-substitutions ()
   "Scan buffer and replace the characters special for Html.
@@ -467,7 +488,7 @@ next to it. Finally, remove or convert all magic entries."
     (goto-char (point-min))
     (while (re-search-forward hyperlatex-meta-all nil t)
       (replace-match
-       (char-to-string (- (preceding-char) hyperlatex-meta-offset))))
+       (char-to-string (- (preceding-char) hyperlatex-meta-offset)) t))
     ;; make labels
     (goto-char (point-min))
     (while (search-forward hyperlatex-meta-X nil t)
@@ -482,7 +503,7 @@ next to it. Finally, remove or convert all magic entries."
       (replace-match (char-to-string
 		      (string-to-number
 		       (buffer-substring (match-beginning 1)
-					 (match-end 1))))))))
+					 (match-end 1)))) t))))
 
 ;;;
 ;;; ----------------------------------------------------------------------
@@ -653,6 +674,29 @@ Deletes command and returns argument (nil if none)."
 	    (buffer-substring (1+ hyperlatex-command-start) (point))
 	  (delete-region hyperlatex-command-start (1+ (point)))))))
 
+(defun hyperlatex-parse-optional-arguments ()
+  "Parses the arguments enclosed in brackets after the commands.
+Returns a list of strings, one for each comma separated option.
+Deletes command and returns argument (nil if none)."
+  (goto-char hyperlatex-command-start)
+  (hyperlatex-delete-|)
+  (hyperlatex-delete-comment)
+  (if (= (following-char) ?\[ )
+      (let ((options nil) (pos (1+ (point))))
+	(goto-char (1+ (point)))
+	(while (/= (following-char) ?\])
+	  (cond
+	    ((= (following-char) ?,)
+		  (setq options (cons (buffer-substring pos (point)) options))
+	      (setq pos (1+ (point))))
+	    ((= (following-char) ?\{)
+	      (forward-sexp 1))
+	  )
+	  (goto-char (1+ (point))))
+	  (add-to-list 'options (buffer-substring pos (point)))
+	  (delete-region hyperlatex-command-start (1+ (point)))
+	  (reverse options))))
+
 (defun hyperlatex-starred-p ()
   "Is current command starred? Remove star, and skip whitespace."
   (hyperlatex-delete-|)
@@ -967,23 +1011,23 @@ Returns NIL if no parameter in string, t otherwise."
   (hyperlatex-insert-required-argument))
 
 (defun hyperlatex-format-input ()
-  (save-excursion
-    (let* ((arg (hyperlatex-parse-required-argument))
-	   (file-name
-	    (cond ((file-readable-p
-		    (expand-file-name arg hyperlatex-input-directory))
-		   (expand-file-name arg hyperlatex-input-directory))
-		  ((file-readable-p (expand-file-name
-				     (concat arg ".tex")
-				     hyperlatex-input-directory))
-		   (expand-file-name (concat arg ".tex")
-				     hyperlatex-input-directory))
-		  (t (error "I can't find the file %s" arg)))))
-      (hyperlatex-insert-file file-name))))
+  (let* ((arg (hyperlatex-parse-required-argument))
+	 (file-name
+	  (cond ((file-readable-p
+		  (expand-file-name arg hyperlatex-input-directory))
+		 (expand-file-name arg hyperlatex-input-directory))
+		((file-readable-p (expand-file-name
+				   (concat arg ".tex")
+				   hyperlatex-input-directory))
+		 (expand-file-name (concat arg ".tex")
+				   hyperlatex-input-directory))
+		(t (error "I can't find the file %s" arg)))))
+    (hyperlatex-insert-file file-name)))
 
 (defun hyperlatex-insert-file (file-name)
   (hyperlatex-message "Inserting file %s..." file-name)
-  (goto-char (+ (point) (car (cdr (insert-file-contents file-name)))))
+  (goto-char (+ (point) 
+		(car (cdr (insert-file-contents-literally file-name)))))
   (narrow-to-region hyperlatex-command-start (point))
   (hyperlatex-prelim-substitutions)
   (goto-char (point-min))
@@ -1000,10 +1044,13 @@ Returns NIL if no parameter in string, t otherwise."
 (put 'htmlpanel		'hyperlatex 'hyperlatex-format-htmlpanel)
 
 (defun hyperlatex-format-xname ()
-  (setq hyperlatex-node-names
-	(cons (cons (1+ hyperlatex-node-number)
-		    (format "%s.html" (hyperlatex-parse-required-argument)))
-	      hyperlatex-node-names)))
+  (if hyperlatex-final-pass
+      (hyperlatex-parse-required-argument)
+    (setq hyperlatex-node-names
+	  (cons (cons (1+ hyperlatex-node-number)
+		      (format "%s%s" (hyperlatex-parse-required-argument)
+			      hyperlatex-html-ext))
+		hyperlatex-node-names))))
 
 (defun hyperlatex-format-htmlpanel ()
   (setq hyperlatex-make-panel
@@ -1103,19 +1150,24 @@ and that we are at the beginning of a new line."
   (setq hyperlatex-made-panel hyperlatex-make-panel)
   (hyperlatex-gen
    (if (= hyperlatex-html-level 32)
-       "!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Draft//EN\""
+       "!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\""
      "!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\"")
    "\n")
   (hyperlatex-gen "HTML" "\n")
   (hyperlatex-gen
    (concat "!-- HTML file produced from " hyperlatex-produced-from
 	   " --\n -- using Hyperlatex v "
-	   hyperlatex-version " (c) Otfried Schwarzkopf --")
+	   hyperlatex-version " (c) Otfried Cheong--\n"
+	   " -- on Emacs " emacs-version ", " (current-time-string) " --")
    "\n")
   (hyperlatex-gen (hyperlatex-get-attributes "HEAD") "\n")
   (hyperlatex-gen (hyperlatex-get-attributes "TITLE"))
   (insert hyperlatex-title (if head (concat " -- " head) ""))
   (hyperlatex-gen "/TITLE" "\n")
+  (if hyperlatex-final-pass
+      (let ((start (point)))
+	(insert "\\HlxMeta{}\n")
+	(hyperlatex-format-region start (point))))
   (hyperlatex-gen "/HEAD")
   (hyperlatex-gen (hyperlatex-get-attributes "BODY") "\n")
   (setq hyperlatex-label-number 1)
@@ -1156,11 +1208,12 @@ and that we are at the beginning of a new line."
 (defun hyperlatex-fullname (node-number)
   (if (and (not hyperlatex-making-frames)
 	   (zerop node-number))
-      (concat hyperlatex-basename ".html")
+      (concat hyperlatex-basename hyperlatex-html-ext)
     (let ((m (assoc node-number hyperlatex-node-names)))
       (if m
 	  (cdr m)
-	(format "%s_%d.html" hyperlatex-basename node-number)))))
+	(format "%s_%d%s" hyperlatex-basename node-number
+		hyperlatex-html-ext)))))
 
 ;;;
 ;;; ----------------------------------------------------------------------
@@ -1186,14 +1239,14 @@ and that we are at the beginning of a new line."
   (delete-region (point-min) (point))
   (hyperlatex-gen
    (if (= hyperlatex-html-level 32)
-       "!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Draft//EN\""
+       "!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\""
      "!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\"")
    "\n")
   (hyperlatex-gen "HTML" "\n")
   (hyperlatex-gen
    (concat "!-- HTML file produced from " hyperlatex-produced-from
 	   " --\n -- using Hyperlatex v "
-	   hyperlatex-version " (c) Otfried Schwarzkopf --")
+	   hyperlatex-version " (c) Otfried Cheong --")
    "\n")
   (hyperlatex-gen (hyperlatex-get-attributes "HEAD") "\n")
   (hyperlatex-gen (hyperlatex-get-attributes "TITLE"))
@@ -1267,7 +1320,7 @@ and that we are at the beginning of a new line."
     (hyperlatex-final-substitutions)
     (write-region (point-min) (point-max)
 		  (concat hyperlatex-html-directory "/"
-			  hyperlatex-basename ".html"))
+			  hyperlatex-basename hyperlatex-html-ext))
     (goto-char (point-max))))
 
 ;;;
@@ -1285,6 +1338,25 @@ and that we are at the beginning of a new line."
 (put 'HlxBackTitle	'hyperlatex 'hyperlatex-format-hlxbacktitle)
 (put 'HlxForwTitle	'hyperlatex 'hyperlatex-format-hlxforwtitle)
 (put 'HlxNodeNumber	'hyperlatex 'hyperlatex-format-hlxnodenumber)
+(put 'HlxThisUrl	'hyperlatex 'hyperlatex-format-hlxthisurl)
+(put 'HlxThisTitle	'hyperlatex 'hyperlatex-format-hlxthistitle)
+
+(defun hyperlatex-this-node ()
+  "Return (sect-num sect-node sect-head sect-lvl sect-label) of current node."
+  (let ((sp hyperlatex-sections))
+    (while (/= (hyperlatex-sect-num (car sp)) hyperlatex-node-section)
+      (setq sp (cdr sp)))
+    ;; sp points to section
+    (car sp)))
+
+(defun hyperlatex-format-hlxthisurl ()
+  "Return URL of current node."
+  (insert (hyperlatex-fullname hyperlatex-node-number)))
+
+(defun hyperlatex-format-hlxthistitle ()
+  "Return title of current node." 
+  (let ((this (hyperlatex-this-node)))
+    (insert (hyperlatex-sect-head this))))
 
 (defun hyperlatex-format-hlxnodenumber ()
   (insert (int-to-string hyperlatex-node-number)))
@@ -1375,8 +1447,7 @@ and that we are at the beginning of a new line."
       (while (/= (hyperlatex-sect-num (car sp)) sect)
 	(setq sp (cdr sp)))
       ;; sp points to section
-      (while (and sp (= hyperlatex-node-number
-			(hyperlatex-sect-node (car sp))))
+      (while (and sp (= sect (hyperlatex-sect-node (car sp))))
 	(setq sp (cdr sp)))
       (if sp (car sp) ()))))
       
@@ -1386,7 +1457,7 @@ and that we are at the beginning of a new line."
     (while (/= (hyperlatex-sect-num (car sp)) sect)
       (setq sp (cdr sp)))
     ;; sp points to section
-    (while (and sp (= hyperlatex-node-number (hyperlatex-sect-node (car sp))))
+    (while (and sp (= sect (hyperlatex-sect-node (car sp))))
       (setq sp (cdr sp)))
     (if sp (car sp) ())))
       
@@ -1484,6 +1555,7 @@ and that we are at the beginning of a new line."
 	      (setq lastlev newlev)
 	      (hyperlatex-gen (hyperlatex-get-attributes "LI"))
 	      (hyperlatex-gen (format hyperlatex-a-href-format
+				      (hyperlatex-get-attributes "A")
 				      (hyperlatex-gen-url
 				       (hyperlatex-sect-node (car sp))
 				       (hyperlatex-sect-label (car sp))
@@ -1566,8 +1638,9 @@ another label (or section heading), use previous label instead."
 	    (cons (list label number hyperlatex-node-number
 			hyperlatex-current-ref)
 		  hyperlatex-labels))
-      (if (string-match "^[a-zA-Z0-9$_.+!*'(),-]+$" label)
-	  ;; label is a legal URL
+      (if (and (string-match "^[a-zA-Z0-9$_.+!*'(),-]+$" label)
+	       (not (string-match "^[0-9]+$" label)))
+	  ;; label is a legal URL and doesn't clash with internal numbers
 	  (setq hyperlatex-label-strings
 		(cons (cons (cons number hyperlatex-node-number)
 			    label)
@@ -1593,7 +1666,9 @@ the anchor of the label."
 		      (concat "[\\\\" hyperlatex-meta-| "]"))
 		     hyperlatex-node-number)
 		  ""))))
-    (hyperlatex-gen (format hyperlatex-a-href-format url))
+    (hyperlatex-gen (format hyperlatex-a-href-format
+			    (hyperlatex-get-attributes "A")
+			    url))
     (insert text)
     (hyperlatex-gen "/A")
     (goto-char hyperlatex-command-start)))
@@ -1646,32 +1721,89 @@ the anchor of the label."
 (put 'endtabular  'hyperlatex 'hyperlatex-end-tabular)
 (put 'htmltab	  'hyperlatex 'hyperlatex-format-tab)
 (put 'hline	  'hyperlatex 'hyperlatex-format-hline)
-(put 'multicolumn 'hyperlatex 'hyperlatex-format-multicolumn)
 (put 'htmlcaption 'hyperlatex 'hyperlatex-format-htmlcaption)
 (put 'multicolumn 'hyperlatex 'hyperlatex-format-multicolumn)
+(put 'newcolumntype 'hyperlatex 'hyperlatex-format-newcolumntype)
 
-(defun hyperlatex-tabular-posn (string)
+(defun hyperlatex-tabular-posn (str)
   "Check the tabular column descriptor and generate a list of
 align tags CENTER, LEFT, RIGHT."
-  (if (not (string-match "^[|lrc]+$" string))
-      (error "Illegal column descriptor.")
-    (let ((i (length string))
-	  (result nil))
-      (while (> i 0)
-	(setq i (1- i))
-	(let ((chr (elt string i)))
-	  (if (= chr ?|)
-	      ()
-	    (setq result (cons (cond ((= chr ?c) "CENTER")
-				     ((= chr ?l) "LEFT")
-				     ((= chr ?r) "RIGHT"))
-			       result)))))
-      result)))
+  (let ((result nil)
+	(prev nil))
+    (while (not (string= str ""))
+      (let ((chr (elt str 0)))
+	(setq str (substring str 1))
+	(cond
+	 ((= chr ?>)
+	  (let ((n (hyperlatex-tabular-parse-argument str chr)))
+	    (setq prev (substring str 1 (1- n)))
+	    (setq str (substring str n))))
+	 ((= chr ?<)
+	  (let* ((n (hyperlatex-tabular-parse-argument str chr))
+		 (next (substring str 1 (1- n))))
+	    (setq str (substring str n))
+	    (setcdr (cdr (car result)) (list next))))
+	 ((= chr ?l)
+	  (setq result (cons (list "LEFT" prev) result))
+	  (setq prev nil))
+	 ((= chr ?c)
+	  (setq result (cons (list "CENTER" prev) result))
+	  (setq prev nil))
+	 ((= chr ?r)
+	  (setq result (cons (list "RIGHT" prev) result))
+	  (setq prev nil))
+	 (t
+	  ;; look into user declared column types
+	  (let* ((match (assoc chr hyperlatex-tabular-column-types))
+		 (narg (elt match 1))
+		 (arglist ())
+		 (decl (elt match 2))
+		 (hook (elt match 3)))
+	    (if (not match)
+		(error (format "unknown column type: %c" chr))
+	      (if hook
+		  (setq str (funcall hook str match))
+		(while (> narg 0)
+		  (let ((n (hyperlatex-tabular-parse-argument str chr)))
+		    (setq arglist (cons (substring str 1 (1- n)) arglist))
+		    (setq str (substring str n))
+		    (setq narg (1- narg))
+		    )
+		  )
+		(setq arglist (cons nil (nreverse arglist)))
+		(while (string-match "##\\([1-9]\\)" decl)
+		  (setq decl
+			(replace-match
+			 (elt arglist (string-to-number (match-string 1 decl)))
+			 t t decl)))
+		(setq str (concat decl str)))))))))
+    (nreverse result)))
+
+(defun hyperlatex-tabular-parse-argument (str help)
+  (let ((chr (elt str 0))
+	(lvl 1)
+	(len 0))
+    (if (not (= chr ?{)) ;}
+	(error (format "tabular: argument required for %c" help))
+      (setq str (substring str 1))
+      (setq len (1+ len))
+      (while (> lvl 0)
+	(if (eq (string-match "^[^{}]*" str) 0)
+	    (progn
+	      (setq len (+ len (match-end 0)))
+	      (setq str (substring str (match-end 0)))))
+	(setq chr (elt str 0))
+	(if (= chr ?{) ;}
+	    (setq lvl (1+ lvl))
+	  (setq lvl (1- lvl)))
+	(setq str (substring str 1))
+	(setq len (1+ len))))
+    len))
 
 (defun hyperlatex-format-tabular ()
   (hyperlatex-parse-optional-argument)
   (setq hyperlatex-tabular-column-descr
-	(cons (cons 1 (hyperlatex-tabular-posn
+	(cons (cons 0 (hyperlatex-tabular-posn
 		       (hyperlatex-parse-required-argument)))
 	      hyperlatex-tabular-column-descr))
   (hyperlatex-blk)
@@ -1683,9 +1815,15 @@ align tags CENTER, LEFT, RIGHT."
      (format (concat "%s COLSPAN=" hyperlatex-meta-dq "1" hyperlatex-meta-dq
 		     " ALIGN=" hyperlatex-meta-dq "%s" hyperlatex-meta-dq)
 	     (hyperlatex-get-attributes "TD")
-	     (car (cdr (car hyperlatex-tabular-column-descr)))) "\n")))
+	     (hyperlatex-tabular-cell-align)) "\n")
+    ;; put cell text into a group to restore attributes like \bf at its end
+    (hyperlatex-begin-group)
+    ;; put contents of >{...} in front of the cell
+    (hyperlatex-tabular-cell-front)))
 
 (defun hyperlatex-end-tabular ()
+  (hyperlatex-tabular-cell-end)
+  (hyperlatex-end-group)
   (setq hyperlatex-tabular-column-descr
 	(cdr hyperlatex-tabular-column-descr))
   (hyperlatex-blk)
@@ -1701,37 +1839,60 @@ align tags CENTER, LEFT, RIGHT."
     (error "Used Tab character `&' outside of tabular environment."))
   (if (<= hyperlatex-html-level 20)
       ()
+    ;; else
+    (hyperlatex-tabular-cell-end)
+    (hyperlatex-end-group)
     (hyperlatex-blk)
+    (setcar (car hyperlatex-tabular-column-descr)
+	    (1+ (car (car hyperlatex-tabular-column-descr))))
     (hyperlatex-gen "/TD")
     (hyperlatex-gen
      (format (concat "%s COLSPAN=" hyperlatex-meta-dq "1" hyperlatex-meta-dq
 		     " ALIGN=" hyperlatex-meta-dq "%s" hyperlatex-meta-dq)
 	     (hyperlatex-get-attributes "TD")
-	     (nth (car (car hyperlatex-tabular-column-descr))
-		  (cdr (car hyperlatex-tabular-column-descr)))))
-    (setcar (car hyperlatex-tabular-column-descr)
-	    (1+ (car (car hyperlatex-tabular-column-descr))))))
+	     (hyperlatex-tabular-cell-align)))
+    (hyperlatex-begin-group)
+    (hyperlatex-tabular-cell-front)))
+
+(defun hyperlatex-tabular-cell-align ()
+  (car (nth (car (car hyperlatex-tabular-column-descr))
+	    (cdr (car hyperlatex-tabular-column-descr)))))
+
+(defun hyperlatex-tabular-cell-front ()
+  (let ((decl (nth 1 (nth (car (car hyperlatex-tabular-column-descr))
+			  (cdr (car hyperlatex-tabular-column-descr))))))
+    (if decl
+	(insert (hyperlatex-evaluate-string decl)))))
+
+(defun hyperlatex-tabular-cell-end ()
+  (let ((decl (nth 2 (nth (car (car hyperlatex-tabular-column-descr))
+			  (cdr (car hyperlatex-tabular-column-descr))))))
+    (if decl
+	(insert (hyperlatex-evaluate-string decl)))))
 
 (defun hyperlatex-format-tab-\\ ()
   (if (<= hyperlatex-html-level 20)
       (insert "\n")
+    (hyperlatex-tabular-cell-end)
+    (hyperlatex-end-group)
     (hyperlatex-blk)
     (hyperlatex-gen "/TD")
     (hyperlatex-gen "/TR" "\n")
+    (setcar (car hyperlatex-tabular-column-descr) 0)
     (hyperlatex-gen "TR")
     (hyperlatex-gen
      (format (concat "%s COLSPAN=" hyperlatex-meta-dq "1" hyperlatex-meta-dq
 		     " ALIGN=" hyperlatex-meta-dq "%s" hyperlatex-meta-dq)
 	     (hyperlatex-get-attributes "TD")
-	     (car (cdr (car hyperlatex-tabular-column-descr)))) "\n")
-    (setcar (car hyperlatex-tabular-column-descr) 1)))
+	     (hyperlatex-tabular-cell-align)) "\n")
+    (hyperlatex-begin-group)
+    (hyperlatex-tabular-cell-front)))
 
 (defun hyperlatex-format-hline ()
-  (if (<= hyperlatex-html-level 20)
-      (progn
-	(hyperlatex-blk)
-	(hyperlatex-gen "HR"))
-    ()))
+  (if (> hyperlatex-html-level 20)
+      nil
+    (hyperlatex-blk)
+    (hyperlatex-gen "HR")))
 
 (defun hyperlatex-format-htmlcaption ()
   (let ((caption (hyperlatex-parse-required-argument)))
@@ -1747,24 +1908,107 @@ align tags CENTER, LEFT, RIGHT."
 	(goto-char here)))))
 
 (defun hyperlatex-format-multicolumn ()
-  (let ((cols (hyperlatex-parse-required-argument))
-	(posn (hyperlatex-tabular-posn (hyperlatex-parse-required-argument)))
-	(item (hyperlatex-parse-required-argument))
-	(here (point-marker)))
+  (let* ((cols (hyperlatex-parse-required-argument))
+	 (posn (car (hyperlatex-tabular-posn
+		     (hyperlatex-parse-required-argument))))
+	 (item (hyperlatex-parse-required-argument))
+	 (here (point-marker)))
     (if (<= hyperlatex-html-level 20)
 	(error "\\multicolumn not available in Html2")
       (re-search-backward
        (concat "COLSPAN=" hyperlatex-meta-dq "1" hyperlatex-meta-dq
-	       " ALIGN=" hyperlatex-meta-dq "[A-Z]+" hyperlatex-meta-dq))
+	       " ALIGN=" hyperlatex-meta-dq "[A-Za-z]+" hyperlatex-meta-dq))
       (replace-match 
        (format (concat "COLSPAN=" hyperlatex-meta-dq "%s" hyperlatex-meta-dq
 		       " ALIGN=" hyperlatex-meta-dq "%s" hyperlatex-meta-dq)
 	       cols (car posn)))
       (goto-char here)
+      (if (nth 1 posn)
+	  (insert (nth 1 posn) "{}"))
       (insert item)
+      (if (nth 2 posn)
+	  (insert (nth 2 posn) "{}"))
       (goto-char here)
       (set-marker here nil))))
 
+(defun hyperlatex-format-newcolumntype ()
+  (let* ((col (string-to-char (hyperlatex-parse-required-argument)))
+	 (narg (hyperlatex-parse-optional-argument))
+	 (decl (hyperlatex-parse-required-argument)))
+    (setq narg (if narg (string-to-number narg) 0))
+    (hyperlatex-tabular-add-columntype col narg decl)))
+
+(defun hyperlatex-tabular-add-columntype (col narg decl &optional hook)
+  (let* ((match (assoc col hyperlatex-tabular-column-types)))
+    (if match
+	(setcdr match (list narg decl hook))
+      (setq hyperlatex-tabular-column-types
+	    (cons (list col narg decl hook)
+		  hyperlatex-tabular-column-types)))))
+
+;;;
+;;; ----------------------------------------------------------------------
+;;;
+;;; Tabbing environment
+;;;
+
+(put 'tabbing	  'hyperlatex 'hyperlatex-format-tabbing)
+(put 'endtabbing  'hyperlatex 'hyperlatex-end-tabbing)
+(put '>           'hyperlatex 'hyperlatex-format-tab-adv)
+(put '=           'hyperlatex 'hyperlatex-format-tab-adv)
+(put 'kill        'hyperlatex 'hyperlatex-format-kill)
+
+(defun hyperlatex-format-tabbing ()
+  (hyperlatex-blk)
+  (if (<= hyperlatex-html-level 20)
+      (error "tabbing not available in Html2")
+    (hyperlatex-gen "TABLE CELLSPACING=\"1\" CELLPADDING=\"0\"")
+    (hyperlatex-gen "TR" "\n")
+    (hyperlatex-gen "TD NOWRAP COLSPAN=\"1\" ALIGN=\"LEFT\"")))
+
+(defun hyperlatex-end-tabbing ()
+  (hyperlatex-blk)
+  (if (<= hyperlatex-html-level 20)
+      (error "tabbing not available in Html2")
+    (hyperlatex-adjust-colspan)
+    (hyperlatex-gen "/TD")
+    (hyperlatex-gen "/TR")
+    (hyperlatex-gen "/TABLE" "\n")))
+
+(defun hyperlatex-format-tab-adv ()
+  (if (not (hyperlatex-in-stack 'tabbing))
+      (error "Used `\\>' or `\\=' outside of tabbing environment."))
+  (if (<= hyperlatex-html-level 20)
+      nil
+    (hyperlatex-blk)
+    (hyperlatex-gen "/TD")
+    (hyperlatex-gen "TD NOWRAP COLSPAN=\"1\" ALIGN=\"LEFT\"")))
+
+(defun hyperlatex-adjust-colspan ()
+  (let ((here (point-marker)))
+    (search-backward "TD NOWRAP COLSPAN=\"1\" ALIGN=\"LEFT\"")
+    (replace-match "TD NOWRAP COLSPAN=\"99\" ALIGN=\"LEFT\"")
+    (goto-char here)
+    (set-marker here nil)))
+
+(defun hyperlatex-format-tabbing-\\ ()
+  (if (<= hyperlatex-html-level 20)
+	(error "tabbing not available in Html2")
+    (hyperlatex-adjust-colspan)
+    (hyperlatex-gen "/TD")
+    (hyperlatex-gen "/TR" "\n")
+    (hyperlatex-gen "TR" "\n")
+    (hyperlatex-gen "TD NOWRAP COLSPAN=\"1\" ALIGN=\"LEFT\"")))
+
+(defun hyperlatex-format-kill ()
+  (if (<= hyperlatex-html-level 20)
+      (error "tabbing not available in Html2")
+    (let ((here (point-marker)))
+      (re-search-backward(concat hyperlatex-meta-< "TR" hyperlatex-meta->))
+      (delete-region (match-beginning 0) here))
+    (hyperlatex-gen "TR" "\n")
+    (hyperlatex-gen "TD NOWRAP COLSPAN=\"1\" ALIGN=\"LEFT\"")))
+
 ;;;
 ;;; ----------------------------------------------------------------------
 ;;;
@@ -1790,7 +2034,9 @@ and in tabular, where it does something else."
       ()
     (if (hyperlatex-in-stack 'tabular)
 	(hyperlatex-format-tab-\\)
-      (hyperlatex-gen "BR"))))
+      (if (hyperlatex-in-stack 'tabbing)
+	  (hyperlatex-format-tabbing-\\)
+	(hyperlatex-gen "BR")))))
   
 (defun hyperlatex-format-example ()
   (hyperlatex-blk)
@@ -1837,7 +2083,6 @@ and in tabular, where it does something else."
 ;;;
 
 (put 'back	'hyperlatex 'hyperlatex-format-backslash)
-(put 'today	'hyperlatex 'hyperlatex-format-today)
 (put 'textsc	'hyperlatex 'hyperlatex-format-textsc)
 (put 'protect	'hyperlatex 'hyperlatex-format-ignore)
 (put 'noindent	'hyperlatex 'hyperlatex-format-ignore)
@@ -1847,24 +2092,6 @@ and in tabular, where it does something else."
   "replace \\back by \\"
   (insert "\\"))
 
-(defun hyperlatex-format-today ()
-  (let* ((date-string (current-time-string))
-	 (month-alist   '(("Jan" . "January") ("Feb" . "February") 
-			  ("Mar" . "March") ("Apr" . "April")
-			  ("May" . "May") ("Jun" . "June") 
-			  ("Jul" . "July") ("Aug" . "August")
-			  ("Sep" . "September") ("Oct" . "October")
-			  ("Nov" . "November") ("Dec" . "December")))
-	 )
-    (string-match "\\(...\\) \\(...\\) \\(..\\).*\\(19..\\)"
-		  (current-time-string) nil)
-    (insert
-     (concat (cdr (assoc (substring date-string 
-				    (match-beginning 2) (match-end 2))
-			 month-alist))
-	     " " (substring date-string (match-beginning 3) (match-end 3))
-	     ", " (substring date-string (match-beginning 4) (match-end 4))))))
-
 (defun hyperlatex-format-textsc ()
   (insert (upcase (hyperlatex-parse-required-argument)))
   (goto-char hyperlatex-command-start))
@@ -2016,6 +2243,8 @@ and in tabular, where it does something else."
 (put 'endifset   'hyperlatex 'hyperlatex-format-ignore)
 (put 'ifclear    'hyperlatex 'hyperlatex-if-clear)
 (put 'endifclear 'hyperlatex 'hyperlatex-format-ignore)
+(put 'ifoption   'hyperlatex 'hyperlatex-if-option)
+(put 'endifoption 'hyperlatex 'hyperlatex-format-ignore)
 (put 'ifequal    'hyperlatex 'hyperlatex-if-equal)
 (put 'endifequal 'hyperlatex 'hyperlatex-format-ignore)
 (put 'ifhtmllevel    'hyperlatex 'hyperlatex-if-htmllevel)
@@ -2078,6 +2307,22 @@ and in tabular, where it does something else."
       (hyperlatex-delete-whitespace)
       (hyperlatex-pop-stacks))))
 
+(defun hyperlatex-if-option ()
+  (let ((arg1 (hyperlatex-parse-evaluated-argument
+	       (concat "[\\\\" hyperlatex-meta-| "]"))))
+    (if (or (and (string= arg1 "")
+	      (not hyperlatex-document-options)
+	      (not hyperlatex-options))
+	  (member arg1 hyperlatex-document-options)
+	  (member arg1 hyperlatex-options))
+	;; process environment
+	()
+      ;; else skip it
+      (delete-region hyperlatex-command-start
+		     (progn (search-forward "\\end{ifoption}") (point)))
+      (hyperlatex-delete-whitespace)
+      (hyperlatex-pop-stacks))))
+
 (defun hyperlatex-if-htmllevel ()
   (let* ((level (hyperlatex-parse-required-argument))
 	 (match (assoc level hyperlatex-html-levels)))
@@ -2328,12 +2573,13 @@ and in tabular, where it does something else."
 (put 'NotSpecial	'hyperlatex 'hyperlatex-format-notspecial)
 (put 'htmllevel		'hyperlatex 'hyperlatex-format-htmllevel)
 (put 'HlxOptions	'hyperlatex 'hyperlatex-format-hlxoptions)
+(put 'htmltopname	'hyperlatex 'hyperlatex-format-htmltopname)
 
 (put 'document		'hyperlatex 'hyperlatex-format-document)
 (put 'enddocument	'hyperlatex 'hyperlatex-end-document)
 
 (defun hyperlatex-format-documentclass ()
-  (setq hyperlatex-options (hyperlatex-parse-optional-argument))
+  (setq hyperlatex-document-options (hyperlatex-parse-optional-arguments))
   (let ((class (hyperlatex-parse-required-argument)))
     (hyperlatex-package "siteinit")
     (hyperlatex-package "init")
@@ -2341,19 +2587,21 @@ and in tabular, where it does something else."
   (goto-char hyperlatex-command-start))
 
 (defun hyperlatex-format-usepackage ()
-  (setq hyperlatex-options (hyperlatex-parse-optional-argument))
+  (setq hyperlatex-options (hyperlatex-parse-optional-arguments))
   (let ((package (hyperlatex-parse-required-argument)))
     (cond ((string= package "hyperlatex")
 	   (if hyperlatex-options
-	       (hyperlatex-package hyperlatex-options)))
+	       (mapcar (function (lambda (pkg) (hyperlatex-package pkg)))
+		 hyperlatex-options)))
 	  ((memq (intern package) hyperlatex-known-packages))
 	  (t
 	   (hyperlatex-package package)))
     (goto-char hyperlatex-command-start)))
 
 (defun hyperlatex-format-hlxoptions ()
-  (if hyperlatex-options
-      (insert hyperlatex-options)))
+  (if (and hyperlatex-options (not (cdr hyperlatex-options)))
+      (insert (car hyperlatex-options))
+    (error "backward compatibility: \\HlxOptions cannot be used with more than one package option")))
 
 (defun hyperlatex-format-htmltitle ()
   (setq hyperlatex-title
@@ -2388,11 +2636,14 @@ and in tabular, where it does something else."
 	(error "Illegal \\htmllevel"))
     (setq hyperlatex-html-level (car (cdr match)))))
 
+(defun hyperlatex-format-htmltopname ()
+  (setq hyperlatex-topnode-name (hyperlatex-parse-required-argument)))
+
 (defun hyperlatex-format-document ()
   "Begin document environment."
   (hyperlatex-message "Title of work is \"%s\"" hyperlatex-title)
-  (hyperlatex-message "Using filename \"%s/%s.html\"" hyperlatex-html-directory
-	   hyperlatex-basename)
+  (hyperlatex-message "Using filename \"%s/%s%s\"" hyperlatex-html-directory
+	   hyperlatex-basename hyperlatex-html-ext)
   (setcar hyperlatex-group-stack "\\htmlfootnotes{}")
   (setq hyperlatex-node-number 0)
   (setq hyperlatex-sect-number 0)
@@ -2409,7 +2660,7 @@ and in tabular, where it does something else."
   (if hyperlatex-final-pass
       ()
     (setq hyperlatex-sections
-	  (cons (list 0 0 "Top" 0 0) hyperlatex-sections))))
+	  (cons (list 0 0 hyperlatex-topnode-name 0 0) hyperlatex-sections))))
 
 (defun hyperlatex-end-document ()
   (delete-region (point) (point-max))
@@ -2423,7 +2674,8 @@ For Latex package PACKAGE, look for `PACKAGE.hlx' in the directories in
 	(fname (hyperlatex-search-file (concat package ".hlx"))))
     (if (not fname)
 	(hyperlatex-message "Package \"%s\" not found..." package)
-      (goto-char (+ (point) (car (cdr (insert-file-contents fname)))))
+      (goto-char (+ (point)
+		    (car (cdr (insert-file-contents-literally fname)))))
       (hyperlatex-message "Package \"%s\" inserted" package))))
 
 (defun hyperlatex-search-file (fname)
